@@ -1,7 +1,7 @@
-from pyg_base import is_date, ulist, logger, is_str, as_list, get_cache, Dict, dictable, list_instances
+from pyg_base import is_date, ulist, logger, is_primitive, is_str, is_strs, as_list, get_cache, Dict, dictable, list_instances
 from pyg_encoders import cell_root, root_path, pd_read_parquet, pickle_load, pd_read_csv, dictable_decode
 from pyg_npy import pd_read_npy
-from pyg_cell._types import _get_mode, _get_qq, DBS, QQ
+from pyg_cell._types import _get_mode, _get_qq, DBS, QQ, CLSS
 from pyg_cell._cell import cell, cell_clear, cell_item, cell_output, cell_func
 from pyg_cell._dag import get_DAG, get_GAD, descendants
 
@@ -9,6 +9,12 @@ from functools import partial
 # import networkx as nx
 import os
 
+def is_partial(db):
+    return isinstance(db, partial)
+
+def not_partial(db):
+    return not isinstance(db, partial)
+
 _readers = dict(parquet = pd_read_parquet, 
                 pickle = pickle_load, 
                 npy = pd_read_npy, 
@@ -30,9 +36,11 @@ _doc = 'doc'
 
 __all__ = ['db_load', 'db_save', 'db_cell', 'cell_push', 'cell_pull', 'get_cell', 'load_cell', 'get_data', 'load_data']
 
+
 def get_GRAPH():
     return get_cache('GRAPH')
 
+
 def is_pairs(pairs):
     """
     returns a check if the data is pairs of key-value tuples
@@ -66,7 +74,6 @@ class db_cell_func(cell_func):
                             raise ValueError('cell is pointing to %s=%s while all cells must point to %s'%(k,keywords.get(k),v))
         return  super(db_cell_func, self).wrapped(*args, **kwargs)           
 
-
         
 def db_save(value):
     """
@@ -117,7 +124,8 @@ def db_load(value, mode = 0):
         return type(value)(**{k : db_load(v, mode = mode) for k, v in value.items()})
     else:
         return value
-        
+
+
 def _load_asof(table, kwargs, deleted, qq = None):  
     t = table.inc(kwargs)
     live = t
@@ -220,11 +228,10 @@ class db_cell(cell):
 
     def __init__(self, function = None, output = None, db = None, **kwargs):
         if db is not None:
-            if not isinstance(db, partial):
-                raise ValueError('type of db is: %s while db must be a *partial* of a function like mongo_table initializing a mongo cursor' % type(db))
-            non_primitives = {key : value for key, value in db.keywords.items() if not _is_primitive(value)}
-            if len(non_primitives):
-                raise ValueError('partial construction of cell must be from primitive paramters. but these are not: %s'%non_primitives)
+            if is_partial(db):
+                non_primitives = {key : value for key, value in db.keywords.items() if not _is_primitive(value)}
+                if len(non_primitives):
+                    raise ValueError('partial construction of cell must be from primitive paramters. but these are not: %s'%non_primitives)
             super(db_cell, self).__init__(function = function, output = output, db = db, **kwargs)
         else:
             self[_db] = None
@@ -234,7 +241,7 @@ class db_cell(cell):
     
     @property
     def _function(self):
-        if 'db' not in self:
+        if not_partial(self.get(_db)):
             function = self.function if isinstance(self.function, cell_func) else cell_func(self.function)
         else:
             keywords = self.db.keywords
@@ -243,7 +250,17 @@ class db_cell(cell):
                 if k in keywords:
                     env[k] = keywords[k]
             if env:
-                function = self.function if isinstance(self.function, db_cell_func) else db_cell_func(self.function)
+                if isinstance(self.function, db_cell_func):
+                    function  = self.function
+                elif isinstance(self.function, cell_func):
+                    f = self.function
+                    function = db_cell_func(f.function, 
+                                                relabels = f.relabels, 
+                                                unloaded = f.unloaded, 
+                                                unitemized = f.unitemized,
+                                                uncalled = f.uncalled)
+                else:
+                    function = db_cell_func(self.function)
                 function.env = env
             else:
                 function = self.function if isinstance(self.function, cell_func) else cell_func(self.function)                
@@ -251,8 +268,11 @@ class db_cell(cell):
 
     @property
     def _pk(self):
-        if self.get(_db) is None:
+        db = self.get(_db)
+        if db is None:
             return super(db_cell, self)._pk
+        elif is_strs(db):
+            return as_list(db)
         else:
             return self.db.keywords.get(_pk)        
 
@@ -273,9 +293,10 @@ class db_cell(cell):
         tuple
             returns a tuple representing the unique address of the cell.
         """
-        if self.get(_db) is None:
-            return super(db_cell, self)._address
-        db = self.db()
+        db = self.get(_db)
+        if not_partial(db):
+            return super(db_cell, self)._address        
+        db = db()
         db_address = db.address
         db_dict = dict(db_address)
         return db_address + tuple([(key, self.get(key)) for key in db._pk if key not in db_dict])
@@ -291,17 +312,18 @@ class db_cell(cell):
             skeletal reference to the database
 
         """
-        if self.get(_db) is None: 
+        if not_partial(self.get(_db)): 
             return super(db_cell, self)._clear()
         else:
             return self[[_db, _function, _updated] + self.db()._pk] if _updated in self else self[[_db, _function] + self.db()._pk]
 
     def save(self):
-        if self.get(_db) is None:
+        db = self.get(_db)
+        if not_partial(db):
             return super(db_cell, self).save()
         address = self._address
         doc = (self - _deleted)
-        db = self.db()
+        db = db()
         missing = ulist(db._pk) - self.keys()
         if len(missing):
             logger.warning('WARN: document not saved as some keys are missing %s'%missing)
@@ -321,7 +343,7 @@ class db_cell(cell):
         """
         loads a document from the database and updates various keys.
         
-        :Persistency:
+        Example:Persistency:
         -------------
         Since we want to avoid hitting the database, there is a singleton GRAPH, a dict, storing the cells by their address.
         Every time we load/save from/to Mongo, we also update GRAPH.
@@ -332,7 +354,20 @@ class db_cell(cell):
         >>> cell.load(-1).load(0)  # clear GRAPH and load from db
         >>> cell.load([0])     # same thing: clear GRAPH and then load if available
 
-        :Merge of cached cell and calling cell:
+        Example:forcing update on change in function inputs:
+        ---------------------------------------------
+        If the current cell, has inputs that is different from the saved document inputs, we force a recalculation by setting updated = None
+
+        >>> from pyg import *
+        >>> db = partial(sql_table, server = 'DESKTOP-LU5C5QF', table = 'test', db = 'test_db', schema = 'dbo', pk = 'key', doc =True)
+        >>> doc = db_cell(add_, a = 1, b = 2, c = 3, key = 'a', db = db)()
+        >>> same_doc = db_cell(db = db, key = 'a').load()
+        >>> assert same_doc.run() is False
+        >>> new = db_cell(add_, a = 2, b = 2, c = 3, key = 'a', db = db).load()
+        >>> assert new.run() is True
+        >>> assert db_cell(add_, a = 2, b = 2, c = 3, key = 'a', db = db)().data == 4
+        
+        Example:Merge of cached cell and calling cell:
         ----------------------------------------
         Once we load from memory (either MongoDB or GRAPH), we tree_update the cached cell with the new values in the current cell.
         This means that it is next to impossible to actually *delete* keys. If you want to delete keys in a cell/cells in the database, you need to:
@@ -341,32 +376,32 @@ class db_cell(cell):
 
         :Parameters:
         ----------
-        mode : int , dataetime, optional
-            if -1, then does not load and clears the GRAPH
-            if 0, then will load from database if found. If not found, will return original document
-            if 1, then will load, if doc not found, data from files (equivalent to db_cell.load_output)
-            if mode is a date, will return the version alive at that date 
+        mode : int , dataetime, str(s), optional
+            if -1,   then will not load and clears the GRAPH
+            if 0,    then will load from database if found. If not found, will return original document
+            if 1,    then will load, if doc not found, data from files (equivalent to db_cell.load_output)
+            if date, then will return the version alive at that date 
+            if strs, then will exclue these keys from the keys loaded from old document
             The default is 0.
             
-            IF you enclose any of these in a list, then GRAPH is cleared prior to running and the database is called.
     
         :Returns:
         -------
         document
 
         """
-        if self.get(_db) is None:
+        if not_partial(self.get(_db)):
             return super(db_cell, self).load(mode = mode)
-        if isinstance(mode, (list, tuple)):
-            if len(mode) == 0:
-                return self.load()
-            if len(mode) == 1:
-                return self.load(mode[0])
-            if len(mode) == 2 and mode[0] == -1:
-                res = self.load(-1)
-                return res.load(mode[-1])
-            else:
-                raise ValueError('mode can only be of the form [], [mode] or [-1, mode]')
+        # if isinstance(mode, (list, tuple)):
+        #     if len(mode) == 0:
+        #         return self.load()
+        #     if len(mode) == 1:
+        #         return self.load(mode[0])
+        #     if len(mode) == 2 and mode[0] == -1:
+        #         res = self.load(-1)
+        #         return res.load(mode[-1])
+        #     else:
+        #         raise ValueError('mode can only be of the form [], [mode] or [-1, mode]')
         db = self.db(mode = 'w')
         pk = ulist(db._pk)
         missing = pk - self.keys()
@@ -390,7 +425,7 @@ class db_cell(cell):
                     graph[doc._address] = doc
                 except Exception:
                     if mode in (1, True):
-                        return self.load_output()
+                        return self.load_output() # just load outputs if exists
                     else:
                         return self
 
@@ -402,10 +437,15 @@ class db_cell(cell):
                 excluded_keys = (self /  None).keys() - self._output - _updated
             else:
                 excluded_keys = (self /  None).keys()
-            if is_date(mode):
+            if is_strs(mode):
+                excluded_keys += as_list(mode)
+            elif is_date(mode):
                 excluded_keys += [_id]
+            updated_inputs = [k for k, v in self._inputs.items() if is_primitive(v) and k in saved and v is not None and saved[k]!=v]
             update = (saved / None) - excluded_keys
             self.update(update)
+            if len(updated_inputs):
+                self[_updated] = None
         return self        
 
     def load_output(self):
@@ -510,7 +550,7 @@ def cell_pull(nodes, types = cell):
 
 
 
-def _get_cell(table = None, db = None, url = None, server = None, deleted = None, _from_memory = None, doc = None, **kwargs):
+def _get_cell(table = None, db = None, url = None, schema = None, server = None, deleted = None, _from_memory = None, doc = None, **kwargs):
     """
     retrieves a cell from a table in a database based on its key words. In addition, can look at earlier versions using deleted.
     It is important to note that this DOES NOT go through the cache mechanism but goes to the database directly every time.
@@ -553,6 +593,7 @@ def _get_cell(table = None, db = None, url = None, server = None, deleted = None
         params.update({key: value for key, value in dict(db = db, 
                                                          url = url, 
                                                          server = server,
+                                                         schema = schema, 
                                                          deleted = deleted,
                                                          _from_memory = _from_memory
                                                          ).items() if value is not None})
@@ -560,16 +601,22 @@ def _get_cell(table = None, db = None, url = None, server = None, deleted = None
         return _get_cell(**params)
     
     pk = kwargs.pop('pk', None)
+    if pk is None:
+        if isinstance(table, partial):
+            pk = as_list(table().pk)
+        elif isinstance(table, tuple(CLSS.values())):
+            pk = as_list(table.pk)
+        elif hasattr(table, 'pk'):
+            pk = as_list(table.pk)
     if pk is None:
         address = kwargs_address = tuple(sorted(kwargs.items()))
     else:
         pk = sorted(as_list(pk))
         address = kwargs_address = tuple([(key, kwargs.get(key)) for key in pk]) 
     
-    schema = kwargs.pop('schema', None)
     mode = _get_mode(url, server, schema = schema)    
     if table is not None:
-        if isinstance(table, partial):
+        if is_partial(table):
             t = table()
             if table.func == DBS.get('sql'):
                 mode = 'sql'
@@ -577,6 +624,14 @@ def _get_cell(table = None, db = None, url = None, server = None, deleted = None
             elif table.func == DBS.get('mongo'):
                 mode = 'mongo'
                 qq = QQ[mode]
+        elif 'sql' in CLSS and isinstance(table, CLSS['sql']):
+            t = table
+            mode = 'sql'
+            qq = t.table.c
+        elif 'mongo' in CLSS and isinstance(table, CLSS['mongo']):
+            t = table
+            mode = 'mongo'
+            qq = QQ[mode]
         elif (db is not None or mode == 'sql'):
             if mode == 'mongo':
                 t = DBS[mode](db = db, table = table, url = url or server, pk = pk)
@@ -601,7 +656,7 @@ def _get_cell(table = None, db = None, url = None, server = None, deleted = None
         return GRAPH[address]
 
 
-def load_cell(table = None, db = None, url = None, server = None, deleted = None, doc = None, **kwargs):
+def load_cell(table = None, db = None, url = None, schema = None, server = None, deleted = None, doc = None, **kwargs):
     """
     retrieves a cell from a table in a database based on its key words. 
     In addition, can look at earlier versions using deleted.
@@ -637,9 +692,9 @@ def load_cell(table = None, db = None, url = None, server = None, deleted = None
     >>> assert load_cell('test','test', surname = 'brown').name == 'bob'
         
     """
-    return _get_cell(table = table, db = db, url = url, deleted = deleted, server = server, doc = doc, **kwargs)
+    return _get_cell(table = table, db = db, url = url, schema = schema, deleted = deleted, server = server, doc = doc, **kwargs)
 
-def get_docs(table = None, db = None, url = None, server = None, pk = None, cell = 'data', doc = None, **kwargs):
+def get_docs(table = None, db = None, url = None, schema = None, server = None, pk = None, cell = 'data', doc = None, **kwargs):
     """
     retrieves multiple cells from a table
 
@@ -655,7 +710,6 @@ def get_docs(table = None, db = None, url = None, server = None, pk = None, cell
         sql server location. The default is None (default to mongodb). Set to True to map to default sql server configured in cfg['sql_server']
 
     """
-    schema = kwargs.pop('schema', None)
     mode = _get_mode(url, server, schema = schema)
     if mode == 'mongo':
         t = DBS[mode](db = db, table = table, url = url or server, pk = pk).inc(**kwargs)
@@ -664,7 +718,7 @@ def get_docs(table = None, db = None, url = None, server = None, pk = None, cell
     return t.docs(list(kwargs.keys()))
 
     
-def get_cell(table = None, db = None, url = None, server = None, deleted = None, doc = None, **kwargs):
+def get_cell(table = None, db = None, schema = None, url = None, server = None, deleted = None, doc = None, **kwargs):
     """
     unlike load_cell which will get the data from the database by default 
     get cell looks at the in-memory cache to see if the cell exists there.
@@ -698,7 +752,7 @@ def get_cell(table = None, db = None, url = None, server = None, deleted = None,
     >>> assert get_cell('test','test', surname = 'brown').name == 'bob'
     """
     _from_memory = kwargs.pop('_from_memory', True)
-    params = dict(table = table, db = db, url = url, server = server, deleted = deleted)
+    params = dict(table = table, db = db, url = url, server = server, deleted = deleted, schema = schema)
     params.update(kwargs)
     multiples = {k : v for k, v in params.items() if isinstance(v, list)}
     if len(multiples) == 0:
@@ -719,7 +773,8 @@ def get_cell(table = None, db = None, url = None, server = None, deleted = None,
         res[doc or _doc] = docs
         return res
 
-def get_data(table = None, db = None, url = None, server = None, deleted = None, doc = None, **kwargs):
+
+def get_data(table = None, db = None, url = None, schema = None, server = None, deleted = None, doc = None, **kwargs):
     """
     retrieves a cell from a table in a database based on its key words. 
     In addition, can look at earlier versions using deleted.
@@ -754,14 +809,15 @@ def get_data(table = None, db = None, url = None, server = None, deleted = None,
     >>> assert get_data('test','test', surname = 'brown') is None
         
     """  
-    cells = get_cell(table, db = db, url = url, server = server, deleted = deleted, doc = doc, **kwargs)
+    cells = get_cell(table, db = db, url = url, schema = schema, server = server, deleted = deleted, doc = doc, **kwargs)
     if isinstance(cells, dictable):
         cells[doc or _doc] = cell_item(cells[doc or _doc], key = 'data')
         return cells
     else:
         return cell_item(cells, key = 'data')
 
-def load_data(table = None, db = None, url = None, server = None, deleted = None, doc = None, **kwargs):
+
+def load_data(table = None, db = None, url = None, schema = None, server = None, deleted = None, doc = None, **kwargs):
     """
     retrieves a cell from a table in a database based on its key words. 
     In addition, can look at earlier versions using deleted.
@@ -798,5 +854,5 @@ def load_data(table = None, db = None, url = None, server = None, deleted = None
     >>> assert load_data('test','test', surname = 'brown') is None
         
     """    
-    return cell_item(load_cell(table, db = db, url = url, server = server, deleted = deleted, doc = doc, **kwargs), key = 'data')
+    return cell_item(load_cell(table, db = db, url = url, schema = schema, server = server, deleted = deleted, doc = doc, **kwargs), key = 'data')
 
